import collections
import itertools
import warnings

import pandas as pd

from scmcallib.distributions import Distribution


class ParameterSet(object):
    """Hold all the data used to describe how the emulation or calibration is constrained.

    Tune parameters
        These parameters are updated to find an optimal set/distribution of parameters

    Config parameters
        Other parameters which are passed to the SCM. Typically these parameters are used to set the models into a specific
        reproducible configuration.

    Another feature of the ParameterSet is that each run get's a globally unique run number (for the current python
    execution).

    Attributes
    ----------
    state: dict
        The current state. The key-value pairs of state is passed as arguments to any callable config parameters when they are evaluated

    """

    run_counter = itertools.count()

    def __init__(self):
        self.tune_parameters = collections.OrderedDict()
        self._config_parameters = collections.OrderedDict()
        self.state = {}

    def __str__(self):
        s = "Config Parameters:\n"
        for p in self._config_parameters:
            s += "{} = {}\n".format(p, self._config_parameters[p])
        s += "Tuning Parameters:\n"
        for p in self.tune_parameters:
            s += "{} = {}\n".format(p, self.tune_parameters[p])
        return s

    def set_tune(self, name, distribution, x0=None):
        """Define a tune parameter

        These are also know as priors, the prior knowledge of the parameter space. These parameters are allowed to change
        when calibrating or emulating.

        When a parameter set is used to emulate another model, the distribution used may not influence the find
        parameters found depending on how the model is setup.

        Parameters
        ----------
        name :
            Name of the parameter. This is the name which is passed to the SCM instance.
        distribution :
            An instance of a subclass of `scmcallib.distributions.Distribution`. This variable defines
            how the parameter can change which emulating or calibration.
        """
        # TODO: handle x0
        if not isinstance(distribution, Distribution):
            raise ValueError(
                "distributions must inherit scmcallib.distributions.Distribution"
            )

        if name in self.config_parameters:
            raise ValueError("{} is already defined as a config parameter")

        self.tune_parameters[name] = distribution

    def set_config(self, name, value):
        """Define a config parameter

        Config parameters are used to set the model into a specific setup for finding point estimates/distributions. They remain
        (mostly) fixed throughout the model runs. These are typically scalars or strings, but functions can also be passed. These
        functions are evaluated on calls to the `config_parameters` property and are passed the contents of `state`.

        Parameters
        ----------
        name :
            Name of the config parameter. This is the name which is used when the value is passed to the initialiser
            of a SCM instance.
        value :
            string or callable. The callables are passed the contents of `ParameterSet.state` as kwargs when they are
            evaluated in `config_parameters`.
        """
        if name in self.tune_parameters:
            raise ValueError("{} is already defined as a config parameter")
        self._config_parameters[name] = value

    def evaluate(self, count, include_config=False):
        """Evaluate the parameters

        Parameters
        ----------
        count :
            Number of parameter sets to sample
        include_config :
            If True include the fixed parameters in the parameter dataframe (Default value = False)

        Returns
        -------
        pd.DataFrame
            DataFrame which contains a set of parameters generated by drawing from the parameter sets
            specified.

        """
        params = collections.OrderedDict(
            {k: self.tune_parameters[k].evaluate(count) for k in self.tune_parameters}
        )
        if include_config:
            params.update(self.config_parameters)

        return pd.DataFrame(
            params, index=[next(self.run_counter) for _ in range(count)]
        )

    @property
    def names(self):
        return self.tune_parameters.keys()

    @property
    def config_parameters(self):
        """Evaluate the config_parameters

        Returns
        -------
        collections.OrderedDict
            If the config parameters are functions they are also evaluated. Any functions that cannot be evaulated then they are excluded
            from the result and a warning is printed.
        """
        res = collections.OrderedDict()

        for k in self._config_parameters:
            v = self._config_parameters[k]
            if callable(v):
                try:
                    v = v(**self.state)
                except TypeError:
                    warnings.warn(
                        "could not evaluate config parameter `{}`. Check `ParameterSet.state`".format(
                            k
                        )
                    )
                    continue
            res[k] = v
        return res

    @property
    def model_specific_parameters(self):
        """Evaluate the config_parameters

        Returns
        -------
        collections.OrderedDict
            If the config parameters are functions they are also evaluated. Any functions that cannot be evaulated then they are excluded
            from the result and a warning is printed.
        """
        res = collections.OrderedDict()

        for k in self._config_parameters:
            v = self._config_parameters[k]
            if callable(v):
                try:
                    v = v(**self.state)
                except TypeError:
                    warnings.warn(
                        "could not evaluate config parameter `{}`. Check `ParameterSet.state`".format(
                            k
                        )
                    )
                    continue
                res[k] = v
        return res


class KeyedParameterSet(ParameterSet):
    """
    ParameterSet where the items being iterated over are already known

    set_config can also handle dicts which are keyed by the values

    This parameter set should be used in preference to ParameterSet if state is changing. That functionality will be removed
    from ParameterSet in future.
    """

    key = None

    def __init__(self, items, key=None):
        super().__init__()
        if key is not None:
            self.key = key
        self.items = items
        self.state = {self.key: self.items[0]}

    def __str__(self):
        s = "Config Parameters:\n"
        vals = {k: self.get_config_parameters(k) for k in self.items}
        df = pd.DataFrame(vals)
        df.columns.name = self.key
        s += str(df)
        s += "\nTuning Parameters:\n"
        for p in self.tune_parameters:
            s += "{} = {}\n".format(p, self.tune_parameters[p])
        return s

    def set_config(self, name, value):
        if callable(value):
            warnings.warn("Deprecating callable config parameters", DeprecationWarning)
        else:
            super().set_config(name, value)

    set_config.__doc__ = ParameterSet.set_config.__doc__

    def get_config_parameters(self, key=None):
        if key is None:
            key = self.state[self.key]
        res = collections.OrderedDict()

        for k in self._config_parameters:
            v = self._config_parameters[k]
            if isinstance(v, dict):
                try:
                    v = v[key]
                except KeyError:
                    warnings.warn(
                        "could not evaluate config parameter `{}`. Check `ParameterSet.state`".format(
                            k
                        )
                    )
                    continue
            elif callable(v):
                warnings.warn(
                    "Deprecating callable config parameters", DeprecationWarning
                )
                continue
            res[k] = v
        return res

    @property
    def config_parameters(self):
        return self.get_config_parameters()

    config_parameters.__doc__ = ParameterSet.config_parameters.__doc__


class ScenarioParameterSet(KeyedParameterSet):
    key = "scenario"


class ModelParameterSet(KeyedParameterSet):
    key = "model"



from logging import getLogger

import numpy as np
import pymc3 as pm
import theano.tensor as tt

from scmcallib.finder.base import BaseFinder
from scmcallib.utils import prepare_dataframe

logger = getLogger(__name__)


class PyMC3Runner(tt.Op):
    """Runner class which is compatible with pymc3.

    We build a wrapper to define our magicc call as a theano operator.
    This is because pymc3 builds on theano operators, it only understands these. See here
    http://deeplearning.net/software/theano/sandbox/how_to_make_ops.html and https://docs.pymc.io/advanced_theano.html
    """

    otypes = [tt.dvector]

    def __init__(self, calibrator, parameters):
        self.calibrator = calibrator
        self.parameters = parameters
        self.itypes = [p.type for p in parameters]

    def perform(self, node, inputs, outputs):
        """the method that is used when calling the Op"""

        # Extract the params and their values
        params = {n.name: float(i) for n, i in zip(node.inputs, inputs)}

        # Run SCM
        scm_outputs = self.get_scm_values(params)
        outputs[0][0] = np.array(scm_outputs)  # output the log-likelihood

    def __call__(self, *args, **kwargs):
        # Expand out the parameters

        return super(PyMC3Runner, self).__call__(*self.parameters)

    def get_scm_values(self, params):
        """Call the SCM and process the results.

        These results are compared against the target timeseries to determine the likelihood

        Parameters
        ----------
        params : dict of parameters to pass to the scm


        Returns
        -------
        pd.Series

        """
        params.update(self.calibrator.parameter_set.config_parameters)
        try:
            results = self.calibrator.scm.run(params)
        except ValueError:
            # Skip this parameter combination
            logger.exception("Could not run scm")
            return np.inf
        results = prepare_dataframe(results)

        # We need to make sure that the ordering is correct. Both dfs are first reindexed using the same index,
        # which is the intersection of the two indices. Then the columns from the target are extracted from the
        # modelled timeseries (maintaining order). A KeyError is raised if the modelled timeseries doesn't
        # have a column specified in the target timeseries
        target = self.calibrator.target
        ts = results.loc[target.index]
        ts -= ts.loc[slice(*self.calibrator.reference_period)].mean()

        # TODO: use filtering from point_estimate

        return ts[target.columns].values.ravel()


class DistributionFinder(BaseFinder):
    """Calibrates a model to give a posterior distribution set of parameters

    This class provides a wrapper to help calibrate a SCM to observation/model data using Monte Carlo Markov Chains (MCMCs).
     An example of setting up a model to calibrate MAGICC7 to global mean temperature observations. Note that calibration
     is slow, requiring many 1000 runs of the SCM.


    Examples
    --------

    .. code-block:: python

        from scmcallib import ParameterSet, DistributionFinder, Normal
        from scmcallib.scm import AR5IR_SCM

        param_set = ParameterSet()
        param_set.set('a1', Normal(mu=1.5, sd=0.1))

        with AR5IR() as scm:
            calibrator = DistributionFinder(param_set, scm=scm)
            calibrator.set_target(observed=observed_gmt)

            results = calibrator.sample(draws=100, tune=50, chains=4)


    Attributes
    ----------
    sampler: pymc3.Model
        The underlying pymc3 model which performs the calibration

    """

    def __init__(self, parameter_set, scm, **kwargs):
        super(DistributionFinder, self).__init__(parameter_set, **kwargs)
        self.scm = scm
        self._runner = None
        self.trace = None

        self.sampler = self.get_sampler()

    def get_sampler(self):
        """Get the pymc3 sampler (model in pymc3 speak) for self's parameter set

        Returns
        -------
        pymc3.Model
            Model created from the parameters and their distributions from ``self.parameter_set``
        """
        # model = pm.Model(theano_config={'compute_test_value': 'off'})
        with pm.Model() as pymc3_model:
            for name, prior in self.parameter_set.tune_parameters.items():
                pymc3_model.Var(name, prior.to_pymc3())

        return pymc3_model

    def sample(self, **kwargs):
        """Sample the posterior

        Any additional arguments are passed onto `pymc3.sample`

        Parameters
        ----------
        kwargs :
            Additional arguments to pass to pymc3.sample.

        Returns
        -------
        pymc3.backends.base.MultiTrace
            TODO: Refactor to a scmcallib results object

        """
        with self.sampler:
            self.trace = pm.sample(**kwargs)

        return self.trace

    def plot(self):
        if self.trace is None:
            raise ValueError("Distribution has not been sampled yet")

        pm.traceplot(self.trace)

    def set_target(self, observed, distribution=pm.Normal, df_kwargs=None, **kwargs):
        """Define the observational data to be modelled

        Parameters
        ----------
        observed :
            Observational/model data to which the SCM should be calibrated
        distribution :
            The pymc3 Distribution which models the uncertainties of the observations. Should typically be pymc3.Normal.
            (Default value = pm.Normal)
        df_kwargs :
            Options to pass to ``create_iam_dataframe`` if using arraylike values for ``observed``. (Default value = None)
        kwargs :
            Other options to pass to the pymc3 distribution
        """
        df_kwargs = {} if df_kwargs is None else df_kwargs

        if "iter_over" in df_kwargs or "iter_over" in kwargs:
            raise NotImplementedError(
                "iter_over for DistributionFinder.set_target not implemented"
            )

        super(DistributionFinder, self).set_target(observed, **df_kwargs)

        if not issubclass(distribution, pm.Distribution):
            raise ValueError("distribution must be a subclass of pm.Distribution")

        if "mu" in kwargs:
            raise ValueError("Cannot pass a value for mu to DistributionFinder.runner")

        self._runner = PyMC3Runner(
            calibrator=self, parameters=self.sampler.deterministics
        )
        self.sampler.Var(
            "output",
            distribution.dist(mu=self._runner(), **kwargs),
            data=observed.values.ravel(),
        )

    def summary(self):
        return pm.summary(self.trace)



from logging import getLogger

from openscm.scmdataframe import df_append

logger = getLogger(__name__)


class BaseSCM(object):
    """Base class which provides an interface to run various simple climate models when fitting

    Each concrete class must implement the `_run_single` method which runs the model.

    Parameters
    ----------
    kwargs : dict
        Initial configuration parameters for the scm. These values maybe used to perform heavy one-off computation.

    Attributes
    ----------
    name: str
        The name of the simple climate model
    """

    name = "base"

    def __init__(self, **kwargs):
        """
        """
        self.config_parameters = kwargs

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.cleanup()

    def run(self, run_parameters, variables=None, iter_over_value=None):
        """Run the model

        Parameters
        ----------
        variables: list
            If not None, the list of variables which are requested from the model. The model must return at least these variables,
            but it free to return more

        kwargs :
            The parameters to run the model with.

            These parameters are joined with the ``config_parameters`` passed to the initialiser.

        Returns
        -------
        openscm.scmdataframe.ScmDataFrame
            An openScm DataFrame containing time series output

        """

        logger.debug(
            "running model {} with config: {}".format(self.name, run_parameters)
        )
        results = self._run_single(run_parameters, variables=variables)

        # Write the extra metadata to the results table
        if iter_over_value is not None:
            for k in iter_over_value:
                results[k] = iter_over_value[k]

        return results

    def run_multiple(self, runs, variables=None):
        """
        Perform multiple SCM runs

        This may be performed serially or in parallel depending on the SCM used

        Parameters
        ----------
        runs: list of tuples
            The runs to perform. Each run is described by a tuple of a dict containing the parameters to run and a dict which
            to overwrite the metadata from the run. i.e. ``{'scenario': 'RCP8.5'} would overwrite the 'scenario' metadata to be
            'RCP8.5' regardless of what the model outputs.
        variables: list of str
            Only return variables in this list
        Returns
        -------
        `openscm.scmdataframe.ScmDataFrame` containing the data from all runs. The ``run_id`` parameter can be used to differentiate
        the runs
        """
        res = [
            self.run(
                run_parameters, variables=variables, iter_over_value=iter_over_value
            )
            for run_parameters, iter_over_value in runs
        ]
        return df_append(res)

    def _run_single(self, parameters, variables=None):
        """
        Perform a single model run

        Parameters
        ----------
        parameters : dict
            A `dict` containing the run-specific configuration
        variables : list of str
            Only return variables in this list

        Returns
        -------
        `openscm.scmdataframe.ScmDataFrame` containing the data from the run

        """
        raise NotImplementedError

    def start(self):
        pass

    def cleanup(self):
        """
        Cleanup any created resources

        Automatically called when exiting a context manager
        """
        pass



import collections
from datetime import datetime

import matplotlib.gridspec as gridspec
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from cycler import cycler

summary_plot_text = """{}
Run Date: {}
{}
Config Parameters:
{}

Model-Specific Parameters:
{}

Tune Parameters:
{}
"""


class SummaryPlot(object):
    """

    Parameters
    ----------
    target : pd.DataFrame

    params_to_show : list of str
        A list of parameter names to plot
    """

    max_timeseries = 50

    def __init__(self, target, parameter_set):
        self.parameter_set = parameter_set
        self.params_to_show = parameter_set.tune_parameters
        self.target = target
        self._text = None
        self._initialise_plot()
        self._write_summary()
        self._timeseries = []

    def _write_summary(self, **kwargs):
        if self._text:
            self._text.remove()

        # Get a table of model specific parameters
        def _get_model_specific(r):
            self.parameter_set.state = r.to_dict()
            return pd.Series(self.parameter_set.model_specific_parameters)

        if len(self.target.columns.levels) > 2:
            meta_table = self.target.columns.droplevel(
                ("region", "variable")
            ).to_frame()
            model_specific_df = meta_table.apply(_get_model_specific, axis=1)
            model_specific_df = "" if model_specific_df.empty else model_specific_df
        else:
            model_specific_df = ""

        text = summary_plot_text.format(
            "",
            datetime.now().isoformat(),
            "Metric: {}".format(kwargs["metric_value"])
            if "metric_value" in kwargs
            else "",
            "\n".join(
                [
                    "{}: {}".format(k, v)
                    for k, v in self.parameter_set.config_parameters.items()
                ]
            ),
            model_specific_df,
            "\n".join(
                "{}: {}".format(
                    tune_param,
                    kwargs[tune_param.lower()] if tune_param.lower() in kwargs else "",
                )
                for tune_param in self.parameter_set.tune_parameters.keys()
            ),
        )

        self._text = self._fig.text(0.01, 0.02, text, fontsize="x-small")

    def _initialise_plot(self):
        plt.ion()
        nparams = len(self.params_to_show)
        regions = sorted(self.target.columns.get_level_values("region").unique())
        variables = sorted(self.target.columns.get_level_values("variable").unique())

        ncols = 2 + len(regions)  # (meta/GOF) | (params) | n * (region)
        nrows = max(max(nparams, len(variables)), 2)
        self._fig = plt.figure(figsize=(5 * ncols, 3 * nrows))
        gs = gridspec.GridSpec(nrows, ncols, figure=self._fig)

        self._ax_metric = plt.subplot(gs[0, 0])
        self._ax_metric.set_title("Metric Value", fontsize="x-small")

        self._ax_params = {}
        common_ax = None
        for i, param in enumerate(self.params_to_show):
            ax = plt.subplot(gs[i, 1], sharex=common_ax)
            ax.set_title(param, fontsize="x-small")
            self._ax_params[param] = ax
            if common_ax is None:
                common_ax = ax

        self._ax_variables = {}

        cc = cycler(color=["0.5", "0.3", "0.7"]) * cycler(
            linestyle=["-", "--", "-.", ":"],
            dashes=[(None, None), (10, 10), (10, 10), (10, 10)],
        )
        cc_iter = cc()
        self._meta_styles = collections.defaultdict(lambda: next(cc_iter))

        prop_cycle = plt.rcParams["axes.prop_cycle"]
        colors = cycler(color=prop_cycle.by_key()["color"])()
        self._meta_colours = collections.defaultdict(lambda: next(colors))

        # Setup the variable plots
        common_ax = None  # All thevariable axes share their x axis
        for col in self.target.columns:
            # index subplots by unique (variable, regions)
            idx = (col[0], col[1])
            if idx not in self._ax_variables:
                variable_idx = variables.index(col[0])
                region_idx = regions.index(col[1])

                ax = plt.subplot(gs[variable_idx, 2 + region_idx], sharex=common_ax)
                if common_ax is None:
                    common_ax = ax
                ax.set_title(idx, fontsize="x-small")
                self._ax_variables[idx] = ax

        #  Work out the bounds of the target variables
        #  The plots are then set up to span the range of the variable and a little bit extra
        self._ax_variable_bounds = {}
        t = self.target.groupby(level=["variable", "region"], axis=1).apply(
            lambda df: pd.Series(np.quantile(df.values, [0, 1.0]), index=[0, 1])
        )
        for idx in t.columns:
            tmax = t[idx][1]
            tmin = t[idx][0]
            mid = (tmax + tmin) / 2
            half_span = (tmax - tmin) / 2  #
            bounds = (mid - half_span * 1.2, tmax + half_span * 1.2)
            self._ax_variable_bounds[idx] = bounds
            self._ax_variables[idx].set_ylim(*bounds)
            if len(self.target.columns.levels) > 2:
                self._ax_variables[idx].legend(fontsize="xx-small")

        self._fig.tight_layout(h_pad=1.5, rect=(0, 0.03, 1, 0.97))
        self._fig.canvas.draw()
        plt.pause(0.01)

    def add_sample(self, fit, redraw=False, **params):
        """
        Add a sample to the summary plot

        Parameters
        ----------
        fit: pd.DataFrame
            The
        params : Tuning parameters, metrics and run number
        """
        params = {k.lower(): params[k] for k in params}
        run_num = params["run_num"]

        # Plot metric
        self._ax_metric.scatter(run_num, params["metric_value"], color="k", marker="x")

        # plot params
        for p in self._ax_params:
            ax = self._ax_params[p]
            ax.scatter(run_num, params[p.lower()], color="k", marker="x")

        self._timeseries.append(fit)
        if len(self._timeseries) > self.max_timeseries:
            self._timeseries = self._timeseries[-self.max_timeseries :]

        if redraw:
            self.redraw_timeseries()
        self._write_summary(**params)

    def redraw_timeseries(self):
        """
        Update the figure

        Operates in a non-blocking manner, i.e. program execution continues unlike ``plt.show``
        """
        # Clear the plots
        for idx in self._ax_variables:
            ax = self._ax_variables[idx]
            ax.clear()
            ax.set_title(idx)
            ax.set_ylim(*self._ax_variable_bounds[idx])

        def plot_fit(ts, style_dict, prefix=None, **styles):
            for col in ts.columns:
                idx = (col[1], col[2])  # col0 is now run_num
                if idx in self._ax_variables:
                    ax = self._ax_variables[idx]
                    name = col[3:] if len(col) > 3 else None
                    ts[col].plot(
                        ax=ax,
                        **styles,
                        **style_dict[name],
                        label="{}: {}".format(prefix, name) if prefix else "_nolegend_",
                    )

        plot_fit(self._timeseries[-1], self._meta_colours, prefix="Fit", linestyle="--")
        # Plot the old timeseries
        for ts in self._timeseries[:-1]:
            plot_fit(ts, self._meta_styles, linewidth=0.2)

        # plot the target
        for col in self.target.columns:
            idx = (col[0], col[1])
            if idx in self._ax_variables:
                ax = self._ax_variables[idx]
                name = col[2:] if len(col) > 2 else None
                self.target[col].plot(
                    ax=ax, **self._meta_colours[name], label="Target: {}".format(name)
                )
        for idx in self._ax_variables:
            ax = self._ax_variables[idx]
            ax.legend()

        self._fig.canvas.draw()
        plt.pause(0.001)  # needed to redraw the plot

    def save_fig(self, fname, **kwargs):
        """
        Save the figure to file

        Parameters
        ----------
        fname : str or file-like object
            A string containing a filepath or an open file object.
        kwargs
            Other optional arguments passed to matplotlib.figure.Figure.savefig

        """
        self._fig.savefig(fname, **kwargs)



"""mat4py - load and save data in the Matlab (TM) MAT-file format

This module provides the following two functions for loading and saving
data in Matlab (TM) MAT-file format:

    data = loadmat(filename, meta=False)

    savemat(filename, data)

The function ``loadmat`` loads all variables stored in the MAT-file into
a simple Python data structure, using only Python's dict and list
objects. Numeric and cell arrays are converted to row-ordered nested lists.
Arrays are squeezed to eliminate arrays with only one element.
The resulting data structure is composed of simple types that are compatible
with the JSON format.

Python data can be saved to a MAT-file, with the function ``savemat``. Data has
to be structured in the same way as for ``loadmat``, i.e. it should be composed
of simple data types, like dict, list, str, int and float.

The following Matlab data structures/types are not supported:

* Arrays with more than 2 dimensions
* Arrays with complex numbers
* Sparse arrays
* Function arrays
* Object classes
* Anonymous function classes
"""
from .loadmat import loadmat
from .savemat import savemat

__version__ = "0.4.2"
__all__ = ["loadmat", "savemat"]
__license__ = """The MIT License (MIT), Copyright (c) 2011-2018 Nephics AB"""


if __name__ == "__main__":
    import cmd

    cmd.main()



import numpy as np
from bayes_opt import BayesianOptimization


from .base import BaseOptimiser


class OptimiserBayesOpt(BaseOptimiser):
    name = "bayesopt"

    def get_param_bounds(self, parameter_set, include_bounds):
        # TODO: put in utils?
        param_bounds = {}
        for name, p in parameter_set.tune_parameters.items():
            if not include_bounds:
                param_bounds[name] = (-np.inf, np.inf)
                continue

            samples = p.distribution.random(size=10000)
            best_guess = samples.mean()
            param_bounds[name] = (
                p.kwargs.get("lower", best_guess + (best_guess - samples.min()) * 1.05),
                p.kwargs.get("upper", best_guess - (best_guess - samples.max()) * 1.05),
            )

        return param_bounds

    def find_best_fit(
        self, evaluator, parameter_set, include_bounds=True, verbose=0, **kwargs
    ):
        kwargs.setdefault("init_points", 100)
        kwargs.setdefault("n_iter", 10)
        kwargs.setdefault("acq", "ei")
        kwargs.setdefault("xi", 0.02)

        param_bounds = self.get_param_bounds(parameter_set, include_bounds)

        bo = BayesianOptimization(f=evaluator, pbounds=param_bounds, verbose=verbose)

        bo.maximize(**kwargs)

        return self.process_minimisation_results(bo)

    def process_minimisation_results(self, bo):
        bo.x = bo.max["params"]
        # BayesOpt doesn't converge in the same sense I don't think, need to check
        # with Nick
        bo.success = True

        return bo



class BaseOptimiser(object):
    """Base class for various optimising backends"""

    name = "base"

    def find_best_fit(self, evaluator, x0, **kwargs):
        raise NotImplementedError

    def minimise(self, evaluator, x0, **kwargs):
        raise NotImplementedError

    def process_minimisation_results(self, res):
        raise NotImplementedError



"""savemat - save data in the Matlab (TM) MAT-file format

Copyright (c) 2011-2018 Nephics AB
The MIT License (MIT)
"""

__all__ = ["savemat"]


import struct
import sys
import time
import zlib

from collections import Sequence, Mapping
from itertools import chain, tee

try:
    from itertools import izip

    ispy2 = True
except ImportError:
    izip = zip
    basestring = str
    ispy2 = False
from io import BytesIO


# encode a string to bytes and vice versa
asbytes = lambda s: s.encode("latin1")
asstr = lambda b: b.decode("latin1")

# array element data types
etypes = {
    "miINT8": {"n": 1, "fmt": "b"},
    "miUINT8": {"n": 2, "fmt": "B"},
    "miINT16": {"n": 3, "fmt": "h"},
    "miUINT16": {"n": 4, "fmt": "H"},
    "miINT32": {"n": 5, "fmt": "i"},
    "miUINT32": {"n": 6, "fmt": "I"},
    "miSINGLE": {"n": 7, "fmt": "f"},
    "miDOUBLE": {"n": 9, "fmt": "d"},
    "miINT64": {"n": 12, "fmt": "q"},
    "miUINT64": {"n": 13, "fmt": "Q"},
    "miMATRIX": {"n": 14},
    "miCOMPRESSED": {"n": 15},
    "miUTF8": {"n": 16, "fmt": "s"},
    "miUTF16": {"n": 17, "fmt": "s"},
    "miUTF32": {"n": 18, "fmt": "s"},
}

# inverse mapping of etypes
inv_etypes = dict((v["n"], k) for k, v in etypes.items())

# matrix array classes
mclasses = {
    "mxCELL_CLASS": 1,
    "mxSTRUCT_CLASS": 2,
    "mxOBJECT_CLASS": 3,
    "mxCHAR_CLASS": 4,
    "mxSPARSE_CLASS": 5,
    "mxDOUBLE_CLASS": 6,
    "mxSINGLE_CLASS": 7,
    "mxINT8_CLASS": 8,
    "mxUINT8_CLASS": 9,
    "mxINT16_CLASS": 10,
    "mxUINT16_CLASS": 11,
    "mxINT32_CLASS": 12,
    "mxUINT32_CLASS": 13,
    "mxINT64_CLASS": 14,
    "mxUINT64_CLASS": 15,
    "mxFUNCTION_CLASS": 16,
    "mxOPAQUE_CLASS": 17,
    "mxOBJECT_CLASS_FROM_MATRIX_H": 18,
}

# map of numeric array classes to data types
numeric_class_etypes = {
    "mxDOUBLE_CLASS": "miDOUBLE",
    "mxSINGLE_CLASS": "miSINGLE",
    "mxINT8_CLASS": "miINT8",
    "mxUINT8_CLASS": "miUINT8",
    "mxINT16_CLASS": "miINT16",
    "mxUINT16_CLASS": "miUINT16",
    "mxINT32_CLASS": "miINT32",
    "mxUINT32_CLASS": "miUINT32",
    "mxINT64_CLASS": "miINT64",
    "mxUINT64_CLASS": "miUINT64",
}

inv_mclasses = dict((v, k) for k, v in mclasses.items())

# data types that may be used when writing numeric data
compressed_numeric = ["miINT32", "miUINT16", "miINT16", "miUINT8"]


def diff(iterable):
    """Diff elements of a sequence:
    s -> s0 - s1, s1 - s2, s2 - s3, ...
    """
    a, b = tee(iterable)
    next(b, None)
    return (i - j for i, j in izip(a, b))


#
# Utility functions
#


def write_file_header(fd):
    # write file header
    desc = "MATLAB 5.0 MAT-file, created with mat4py on: " + time.strftime(
        "%a, %b %d %Y %H:%M:%S", time.localtime()
    )
    fd.write(struct.pack("116s", desc.encode("latin1")))
    fd.write(struct.pack("8s", b" " * 8))
    fd.write(struct.pack("H", 0x100))
    if sys.byteorder == "big":
        fd.write(struct.pack("2s", b"MI"))
    else:
        fd.write(struct.pack("2s", b"IM"))


def write_elements(fd, mtp, data, is_name=False):
    """Write data element tag and data.

    The tag contains the array type and the number of
    bytes the array data will occupy when written to file.

    If data occupies 4 bytes or less, it is written immediately
    as a Small Data Element (SDE).
    """
    fmt = etypes[mtp]["fmt"]
    if isinstance(data, Sequence):
        if fmt == "s" or is_name:
            if isinstance(data, bytes):
                if is_name and len(data) > 31:
                    raise ValueError(
                        'Name "{}" is too long (max. 31 '
                        "characters allowed)".format(data)
                    )
                fmt = "{}s".format(len(data))
                data = (data,)
            else:
                fmt = "".join("{}s".format(len(s)) for s in data)
        else:
            l = len(data)
            if l == 0:
                # empty array
                fmt = ""
            if l > 1:
                # more than one element to be written
                fmt = "{}{}".format(l, fmt)
    else:
        data = (data,)
    num_bytes = struct.calcsize(fmt)
    if num_bytes <= 4:
        # write SDE
        if num_bytes < 4:
            # add pad bytes
            fmt += "{}x".format(4 - num_bytes)
        fd.write(struct.pack("hh" + fmt, etypes[mtp]["n"], *chain([num_bytes], data)))
        return
    # write tag: element type and number of bytes
    fd.write(struct.pack("b3xI", etypes[mtp]["n"], num_bytes))
    # add pad bytes to fmt, if needed
    mod8 = num_bytes % 8
    if mod8:
        fmt += "{}x".format(8 - mod8)
    # write data
    fd.write(struct.pack(fmt, *data))


def write_var_header(fd, header):
    """Write variable header"""

    # write tag bytes,
    # and array flags + class and nzmax (null bytes)
    fd.write(struct.pack("b3xI", etypes["miUINT32"]["n"], 8))
    fd.write(struct.pack("b3x4x", mclasses[header["mclass"]]))

    # write dimensions array
    write_elements(fd, "miINT32", header["dims"])

    # write var name
    write_elements(fd, "miINT8", asbytes(header["name"]), is_name=True)


def write_var_data(fd, data):
    """Write variable data to file"""
    # write array data elements (size info)
    fd.write(struct.pack("b3xI", etypes["miMATRIX"]["n"], len(data)))

    # write the data
    fd.write(data)


def write_compressed_var_array(fd, array, name):
    """Write compressed variable data to file"""
    bd = BytesIO()

    write_var_array(bd, array, name)

    data = zlib.compress(bd.getvalue())
    bd.close()

    # write array data elements (size info)
    fd.write(struct.pack("b3xI", etypes["miCOMPRESSED"]["n"], len(data)))

    # write the compressed data
    fd.write(data)


def write_numeric_array(fd, header, array):
    """Write the numeric array"""
    # make a memory file for writing array data
    bd = BytesIO()

    # write matrix header to memory file
    write_var_header(bd, header)

    if not isinstance(array, basestring) and header["dims"][0] > 1:
        # list array data in column major order
        array = list(chain.from_iterable(izip(*array)))

    # write matrix data to memory file
    write_elements(bd, header["mtp"], array)

    # write the variable to disk file
    data = bd.getvalue()
    bd.close()
    write_var_data(fd, data)


def write_cell_array(fd, header, array):
    # make a memory file for writing array data
    bd = BytesIO()

    # write matrix header to memory file
    write_var_header(bd, header)

    for row in range(header["dims"][0]):
        for col in range(header["dims"][1]):
            if header["dims"][0] > 1:
                vdata = array[row][col]
            else:
                # array is squeezed on the row dimension
                vdata = array[col]
            write_var_array(bd, vdata)

    # write the variable to disk file
    data = bd.getvalue()
    bd.close()
    write_var_data(fd, data)


def write_struct_array(fd, header, array):
    # make a memory file for writing array data
    bd = BytesIO()

    # write matrix header to memory file
    write_var_header(bd, header)

    fieldnames = list(array.keys())

    field_names_sizes = [(f, len(f)) for f in (asbytes(s) for s in fieldnames)]

    # write field name length (the str length + a null byte)
    field_length = max(i[1] for i in field_names_sizes) + 1
    if field_length > 32:
        raise ValueError("Struct field names are too long")
    write_elements(bd, "miINT32", field_length)

    # write fieldnames
    write_elements(
        bd,
        "miINT8",
        [f + (field_length - l % field_length) * b"\0" for f, l in field_names_sizes],
        is_name=True,
    )

    # wrap each field in a cell
    for row in range(header["dims"][0]):
        for col in range(header["dims"][1]):
            for field in fieldnames:
                if header["dims"][0] > 1:
                    vdata = array[field][col][row]
                elif header["dims"][1] > 1:
                    vdata = array[field][col]
                else:
                    vdata = array[field]
                write_var_array(bd, vdata)

    # write the variable to disk file
    data = bd.getvalue()
    bd.close()
    write_var_data(fd, data)


def write_char_array(fd, header, array):
    if isinstance(array, basestring):
        # split string into chars
        array = [asbytes(c) for c in array]
    else:
        # split each string in list into chars
        array = [[asbytes(c) for c in s] for s in array]
    return write_numeric_array(fd, header, array)


def write_var_array(fd, array, name=""):
    """Write variable array (of any supported type)"""
    header, array = guess_header(array, name)
    mc = header["mclass"]
    if mc in numeric_class_etypes:
        return write_numeric_array(fd, header, array)
    elif mc == "mxCHAR_CLASS":
        return write_char_array(fd, header, array)
    elif mc == "mxCELL_CLASS":
        return write_cell_array(fd, header, array)
    elif mc == "mxSTRUCT_CLASS":
        return write_struct_array(fd, header, array)
    else:
        raise ValueError("Unknown mclass {}".format(mc))


def isarray(array, test, dim=2):
    """Returns True if test is True for all array elements.
    Otherwise, returns False.
    """
    if dim > 1:
        return all(isarray(array[i], test, dim - 1) for i in range(len(array)))
    return all(test(i) for i in array)


def guess_header(array, name=""):
    """Guess the array header information.
    Returns a header dict, with class, data type, and size information.
    """
    header = {}

    if isinstance(array, Sequence) and len(array) == 1:
        # sequence with only one element, assume user wants cell
        header.update({"mclass": "mxCELL_CLASS", "dims": (1, 1)})

    elif isinstance(array, basestring):
        header.update(
            {
                "mclass": "mxCHAR_CLASS",
                "mtp": "miUTF8",
                "dims": (1 if len(array) > 0 else 0, len(array)),
            }
        )

    elif isinstance(array, Sequence) and len(array) == 0:
        # empty (int) array
        header.update({"mclass": "mxINT32_CLASS", "mtp": "miINT32", "dims": (0, 0)})

    elif isinstance(array, Mapping):
        # test if cells (values) of all fields are of equal type and
        # have equal length
        field_types = [type(j) for j in array.values()]
        field_lengths = [
            1 if isinstance(j, (basestring, int, float)) else len(j)
            for j in array.values()
        ]
        if len(field_lengths) == 1:
            equal_lengths = True
            equal_types = True
        else:
            equal_lengths = not any(diff(field_lengths))
            equal_types = all([field_types[0] == f for f in field_types])

        # if of unqeual lengths or unequal types, then treat each value
        # as a cell in a 1x1 struct
        header.update(
            {
                "mclass": "mxSTRUCT_CLASS",
                "dims": (1, field_lengths[0] if equal_lengths and equal_types else 1),
            }
        )

    elif isinstance(array, int):
        header.update({"mclass": "mxINT32_CLASS", "mtp": "miINT32", "dims": (1, 1)})

    elif isinstance(array, float):
        header.update({"mclass": "mxDOUBLE_CLASS", "mtp": "miDOUBLE", "dims": (1, 1)})

    elif isinstance(array, Sequence):

        if isarray(array, lambda i: isinstance(i, int), 1):
            # 1D int array
            header.update(
                {"mclass": "mxINT32_CLASS", "mtp": "miINT32", "dims": (1, len(array))}
            )

        elif isarray(array, lambda i: isinstance(i, (int, float)), 1):
            # 1D double array
            header.update(
                {"mclass": "mxDOUBLE_CLASS", "mtp": "miDOUBLE", "dims": (1, len(array))}
            )

        elif isarray(array, lambda i: isinstance(i, Sequence), 1) and any(
            diff(len(s) for s in array)
        ):
            # sequence of unequal length, assume cell array
            header.update({"mclass": "mxCELL_CLASS", "dims": (1, len(array))})

        elif isarray(array, lambda i: isinstance(i, basestring), 1):
            # char array
            header.update(
                {
                    "mclass": "mxCHAR_CLASS",
                    "mtp": "miUTF8",
                    "dims": (len(array), len(array[0])),
                }
            )

        elif isarray(array, lambda i: isinstance(i, Sequence), 1):
            # 2D array

            if any(diff(len(j) for j in array)):
                # rows are of unequal length, make it a cell array
                header.update(
                    {"mclass": "mxCELL_CLASS", "dims": (len(array), len(array[0]))}
                )

            elif isarray(array, lambda i: isinstance(i, int)):
                # 2D int array
                header.update(
                    {
                        "mclass": "mxINT32_CLASS",
                        "mtp": "miINT32",
                        "dims": (len(array), len(array[0])),
                    }
                )

            elif isarray(array, lambda i: isinstance(i, (int, float))):
                # 2D double array
                header.update(
                    {
                        "mclass": "mxDOUBLE_CLASS",
                        "mtp": "miDOUBLE",
                        "dims": (len(array), len(array[0])),
                    }
                )

        elif isarray(
            array, lambda i: isinstance(i, (int, float, basestring, Sequence, Mapping))
        ):
            # mixed contents, make it a cell array
            header.update({"mclass": "mxCELL_CLASS", "dims": (1, len(array))})

    if not header:
        raise ValueError(
            "Only dicts, two dimensional numeric, "
            "and char arrays are currently supported"
        )
    header["name"] = name
    return header, array


#
# Write to MAT file
#


def savemat(filename, data):
    """Save data to MAT-file:

    savemat(filename, data)

    The filename argument is either a string with the filename, or
    a file like object.

    The parameter ``data`` shall be a dict with the variables.

    A ``ValueError`` exception is raised if data has invalid format, or if the
    data structure cannot be mapped to a known MAT array type.
    """

    if not isinstance(data, Mapping):
        raise ValueError("Data should be a dict of variable arrays")

    if isinstance(filename, basestring):
        fd = open(filename, "wb")
    else:
        fd = filename

    write_file_header(fd)

    # write variables
    for name, array in data.items():
        write_compressed_var_array(fd, array, name)

    fd.close()



import collections
from copy import deepcopy
from logging import getLogger

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from openscm.scmdataframe import ScmDataFrame
from openscm.scmdataframe.base import REQUIRED_COLS, ScmDataFrameBase
from scipy.interpolate import griddata

from scmcallib.finder.base import BaseFinder
from scmcallib.metrics import get_metric, metrics, LARGE_NEG_ERR
from scmcallib.optimiser import get_optimiser
from scmcallib.plot import SummaryPlot
from scmcallib.utils import prepare_dataframe

logger = getLogger(__name__)


class EmulationResult(object):
    """Holds the results from emulating some predictor timeseries

    Contains the output coefficients and a number of plotting routines

    """

    def __init__(self, result, target, evaluator):
        # Check that the emulator converged
        assert result.success, "The emulator failed to converge: {}".format(
            result.message
        )

        self._result = result
        self.target = target
        self.parameter_set = evaluator.parameter_set
        self.samples = evaluator.samples
        self.scm_df = evaluator.scm_df
        self.metrics = evaluator.metrics(**self.coeffs)

    def __len__(self):
        return len(self.samples)

    @property
    def coeffs(self):
        return self._result.x

    def plot_summary(self, fname=None, **kwargs):
        s = SummaryPlot(self.target, self.parameter_set)

        # Plot metric
        df = self.scm_df
        metric_vals = df[["run_num", "metric_value"]].drop_duplicates()
        s._ax_metric.scatter(
            metric_vals["run_num"], metric_vals["metric_value"], color="k", marker="x"
        )

        # plot params
        for p in s._ax_params:
            ax = s._ax_params[p]
            d = df[["run_num", p.lower()]]
            ax.scatter(d["run_num"], d[p.lower()], color="k", marker="x")

        # Queue up the timeseries
        for run_num in df["run_num"].drop_duplicates():
            ts = (
                df.filter(run_num=run_num)
                .timeseries(["run_num"] + self.target.columns.names)
                .T
            )

            s._timeseries.append(ts)
            if len(s._timeseries) > s.max_timeseries:
                s._timeseries = s._timeseries[-s.max_timeseries :]

        s.redraw_timeseries()
        s._write_summary(**df.meta.iloc[-1].to_dict())

        if fname is not None:
            s.save_fig(fname, **kwargs)

    def plot_fit(self, **plt_kwargs):
        nrows = len(self.target.columns)
        fig, axes = plt.subplots(nrows=nrows, **plt_kwargs)
        if nrows == 1:
            axes = [axes]

        for i, v in enumerate(self.target):
            filter = {k: val for k, val in zip(self.target.columns.names, v)}
            self.target[v].plot(ax=axes[i])
            self.fit.filter(**filter).timeseries().T.plot(
                ax=axes[i]
            )  # fit is a ScmDataFrame
            axes[i].set_title(v)
            axes[i].legend(["Obs", "Fit"])
        plt.tight_layout()
        return fig, axes

    def plot_heatmap(self, v1, v2, gridsize=15, **kwargs):
        self.samples.plot.hexbin(v1, v2, gridsize=gridsize, **kwargs)
        plt.gca().plot([self.coeffs[v1]], [self.coeffs[v2]], "rx")

    def plot_scatter(self, v1, v2, **kwargs):
        self.samples.plot.scatter(v1, v2, **kwargs)

    def plot_likelihood_surface(self, v1, v2, **kwargs):
        fig, ax = plt.subplots()
        x = self.samples[v1].astype(float)
        y = self.samples[v2].astype(float)

        metric_values = self.samples["metric_value"].astype(float).values.reshape(-1, 1)

        m = 200
        n = 200

        # define grid
        x_lin = np.linspace(x.min() * 1.2, x.max() * 0.9, m)
        y_lin = np.linspace(y.min() * 1.2, y.max() * 0.9, n)
        x_grid, y_grid = np.meshgrid(x_lin, y_lin)

        # grid the data
        z_grid = griddata(
            points=(x, y), values=metric_values, xi=(x_grid, y_grid), method="linear"
        )
        x_grid = x_grid.reshape(m, n)
        y_grid = y_grid.reshape(m, n)
        z_grid = z_grid.reshape(m, n)
        # contour the gridded data
        levels = np.logspace(0, 5, 20)

        plt.contour(
            x_grid, y_grid, z_grid, norm=matplotlib.colors.LogNorm(), levels=levels
        )

        plt.colorbar()  # draw colorbar
        # plot data points.
        ax.set_xlim(x_grid.min() * 1.2, x_grid.max() * 0.9)
        ax.set_ylim(y_grid.min() * 1.2, y_grid.max() * 0.9)
        ax.set_xlabel(v1)
        ax.set_ylabel(v2)

        ax.plot(x, y, "bo")
        ax.plot([self.coeffs[v1]], [self.coeffs[v2]], "rx")

    def _get_grid_bounds(x, y, x_trim, y_trim):

        x_range = x.max() - x.min()
        y_range = y.max() - y.min()

        x_min = x.min() + 0.5 * x_trim * x_range
        x_max = x.max() - 0.5 * x_trim * x_range

        y_min = y.min() + 0.5 * y_trim * y_range
        y_max = y.max() - 0.5 * y_trim * y_range

        return x_min, x_max, y_min, y_max

    @property
    def fit(self):
        # Assume that the last run is the final run
        run_idx = self.samples.iloc[-1].name
        return self.scm_df.filter(run_num=run_idx)

    def print_summary(self):
        print(self._result)


class Evaluator(object):
    def __init__(self, fitter, scm):
        self.fitter = fitter
        self.parameter_set = fitter.parameter_set
        self.parameter_names = fitter.parameter_set.names
        self.metric_func = fitter.metric_func
        self.scm = scm
        self.target = fitter.target

        # Storage for coefficients
        self._samples = collections.defaultdict(list)

        self._timeseries = []
        self.prev_fit = None

        self.live_plot = fitter.live_plot
        self.sample_step = fitter.sample_step
        if self.live_plot:
            self._summary_plot = SummaryPlot(self.target, self.parameter_set)

    def __call__(self, run_num=None, force_add=False, **run_parameters):
        """Runs the SCM with a given configuration and compare the results against the target data

        Parameters
        ----------
        run_num: int
            Optional value which uniquely identifies the run

        run_parameters: dict
            SCM parameters to run

        Returns
        -------
        float
            The metric_value calculated using self.metric_func
        """

        variables = list(self.target.columns.get_level_values("variable").unique())
        if self.fitter.iter_over_values is not None:
            all_cfg = []
            for v in self.fitter.iter_over_values:
                self.parameter_set.state = v
                all_cfg.append(
                    ({**run_parameters, **self.parameter_set.config_parameters}, v)
                )
            results = self.scm.run_multiple(all_cfg, variables=variables)
        else:
            results = self.scm.run(run_parameters, variables=variables)

        run_num = (
            run_num if run_num is not None else next(self.parameter_set.run_counter)
        )
        results = prepare_dataframe(results, self.fitter.iter_over)

        assert len(results), "Model returned no valid data"

        # Do the merging manually for now
        # results = self.scm.process_results(results)
        merged_idx = results.index.intersection(self.target.index)

        if len(merged_idx) == 0:
            raise ValueError("Output did not overlap with the target timeseries")
        ts = results.loc[merged_idx]
        ts -= ts.loc[slice(*self.fitter.reference_period)].mean()

        try:
            # We need to make sure that the ordering is correct. Both dfs are first use the same index which is the intersection of
            # the two indices. Then the columns from the target are extracted from the modelled timeseries (maintaining order).
            # A KeyError is raised if the modelled timeseries doesn't have a column specified in the target timeseries
            scm_ts = ts[self.target.columns].values.ravel()
            target_ts = self.target.loc[merged_idx].values.ravel()
            assert scm_ts.shape == target_ts.shape
            mask = np.isnan(scm_ts) | np.isnan(target_ts)

            if np.sum(~mask) == 0:
                raise ValueError("Could not find any overlapping non-nan values")

            weights = None
            if self.fitter.weights is not None:
                weights = self.fitter.weights.loc[merged_idx][
                    self.target.columns
                ].values.ravel()[~mask]

            metric_value = self.metric_func(
                scm_ts[~mask], target_ts[~mask], sample_weight=weights
            )
        except KeyError:
            logger.exception(
                "not all target data available. Have {}".format(ts.columns.to_list())
            )
            raise ValueError(
                "could not find overlapping scm time series for all of the training time series"
            )
        except ValueError:
            logger.warning(
                "Could not calculate metric value for run {}".format(run_num)
            )
            metric_value = np.nan

        if not np.isfinite(metric_value):
            logger.warning('Non finite metric value found. using LARGE_NEG_ERR ({:g})'.format(LARGE_NEG_ERR))
            metric_value = LARGE_NEG_ERR

        to_save = run_parameters.copy()

        col_names = ts.columns.names
        ts = ts.T.reset_index()
        ts["run_num"] = run_num
        ts = ts.set_index(["run_num"] + col_names).T
        self.prev_fit = ts

        if force_add or (
            self.sample_step is not None and run_num % self.sample_step == 0
        ):
            to_save["metric_value"] = metric_value
            to_save["run_num"] = run_num
            self._timeseries.append(ts)
            for k in to_save:
                self._samples[k].append(to_save[k])

            if self.live_plot:
                self._summary_plot.add_sample(ts, **to_save)
                self._summary_plot.redraw_timeseries()

            logger.info("run_num: {} metric_value: {}".format(run_num, metric_value))
        return metric_value

    def run_multiple(self, df):
        """Takes a df and performs several runs

        Runs several runs of the scm calulating the metric_value on each iteration

        Parameters
        ----------
        df : pd.DataFrame
            DataFrame containing the parameters to run. This can be easily obtained using `ParameterSet.evaluate`.

        Returns
        -------
        pd.DataFrame
            The ``df`` dataframe with an additional ``metric_value`` column


        See Also
        --------
        ``ParameterSet.evaluate``

        """
        res = [
            self(**coeffs.to_dict(), run_num=run_num)
            for run_num, coeffs in df.iterrows()
        ]
        df["metric_value"] = pd.Series(res, index=df.index)

        return df

    @property
    def samples(self):
        """Get a dataframe containing all of the sampled points during the life of the Evaluator

        Only create the DataFrame when needed as appending to a DF a row at a time is very slow

        Returns
        -------
        pd.DataFrame

        """
        d = self._samples.copy()
        if len(d["run_num"]) == 0:
            return pd.DataFrame()

        run_num = d.pop("run_num")
        df = pd.DataFrame(d, index=run_num)
        df.index.name = "run_num"
        return df

    @property
    def scm_df(self):
        """Get a dataframe containing all of the sampled points during the life of the Evaluator

        Only create the DataFrame when needed as appending to a DF a row at a time is very slow

        Returns
        -------
        openscm.dataframe.ScmDataFrame
        """

        if not len(self._timeseries):
            if self.prev_fit is None:
                return None
            self._timeseries = [self.prev_fit]

        from functools import reduce, partial

        outer_merge = partial(pd.merge, how="outer", left_index=True, right_index=True)
        df = reduce(outer_merge, self._timeseries)

        # merge in the extra metadata from the target timeseries
        if self.fitter.extra_meta is not None:
            meta_fields = self.fitter.extra_meta.index.names
            df = pd.merge(
                df.T.reset_index(),
                self.fitter.extra_meta.reset_index(),
                left_on=meta_fields,
                right_on=meta_fields,
            )
        else:
            df = df.T.reset_index()
        # Finally merge in the parameter values
        df = pd.merge(df, self.samples, left_on="run_num", right_index=True)
        df["climate_model"] = self.scm.name
        for f in REQUIRED_COLS:
            if f not in df:
                df[f] = "Unknown"
        df = ScmDataFrame(df)

        return df

    def metrics(self, **run_parameters):
        """Calculates all of the metrics for a given run configuration

        Typically only called on the fit

        Returns
        -------
        dict containing all the calculated metrics
        """

        variables = list(self.target.columns.get_level_values("variable").unique())
        if self.fitter.iter_over_values is not None:
            all_cfg = []
            for v in self.fitter.iter_over_values:
                self.parameter_set.state = v
                all_cfg.append(
                    ({**run_parameters, **self.parameter_set.config_parameters}, v)
                )
            results = self.scm.run_multiple(all_cfg, variables=variables)
        else:
            results = self.scm.run(run_parameters, variables=variables)

        results = prepare_dataframe(results, self.fitter.iter_over)

        assert len(results), "Model returned no valid data"

        # Do the merging manually for now
        # results = self.scm.process_results(results)
        merged_idx = results.index.intersection(self.target.index)

        if len(merged_idx) == 0:
            raise ValueError("Output did not overlap with the target timeseries")
        ts = results.loc[merged_idx]
        ts -= ts.loc[slice(*self.fitter.reference_period)].mean()

        res = {}
        try:
            # We need to make sure that the ordering is correct. Both dfs are first use the same index which is the intersection of
            # the two indices. Then the columns from the target are extracted from the modelled timeseries (maintaining order).
            # A KeyError is raised if the modelled timeseries doesn't have a column specified in the target timeseries
            scm_ts = ts[self.target.columns].values.ravel()
            target_ts = self.target.loc[merged_idx].values.ravel()
            assert scm_ts.shape == target_ts.shape
            mask = np.isnan(scm_ts) | np.isnan(target_ts)

            if np.sum(~mask) == 0:
                raise ValueError("Could not find any overlapping non-nan values")

            weights = None
            if self.fitter.weights is not None:
                weights = self.fitter.weights.loc[merged_idx][
                    self.target.columns
                ].values.ravel()[~mask]

            for m in metrics:
                try:
                    res[m] = metrics[m](
                        scm_ts[~mask], target_ts[~mask], sample_weight=weights
                    )
                except ValueError:
                    logger.warning(
                        "Could not calculate metric value"
                    )
                    res[m] = np.nan
        except KeyError:
            logger.exception(
                "not all target data available. Have {}".format(ts.columns.to_list())
            )
            raise ValueError(
                "could not find overlapping scm time series for all of the training time series"
            )
        return res


class PointEstimateFinder(BaseFinder):
    """Generate a set of best-fit parameters which best emulate a given timeseries

    This class provides a wrapper to help emulate observational or model data to produce a single set of parameters.

    An example of setting up a model to calibrate MAGICC7 to global mean temperature observations

    .. code-block:: python

        emulator = PointEstimateFinder(param_set)
        emulator.set_target(observations)

        with MAGICC7_SCM() as scm:
            results = emulator.find_best_fit(scm)

    Parameters
    ----------
    parameter_set : scmcallib.parameterset.ParameterSet
        An instance of scmcallib.parameterset.ParameterSet which is used to define the constraints
        on the model. The config parameters will be updated to find the combination of parameters which best match the
        target data set using `set_target`.
    metric : str
        The metric used to calculate how well the modelled data matches the observations. This metric is
        the variable maximised in `find_best_fit`. Options include:

        - r_squared: R^2 value
        - explained_variance: Measure of the extent to which the modelled data accounts for the variation
        - neg_mean_sq_err: Mean square error (MSE)
        - neg_mean_abs_err: Mean absolute error (MAE)
        - neg_mean_sq_log_err: Mean square log error (MSLE)
        - neg_median_abs_err: Median absolute error (MedAE)
    sample_step : int or None
        Used to determine how often the parameter sets and fits are saved. ``sample_step`` is the step size between saving so for example
        a ``sample_step`` of 1 means that every sample is recorded while for a ``sample_step`` is 5 then every 5th sample is recorded.
        If ``sample_step`` is None then only the most recent sample is stored in the ``prev_fit`` attribute.
    live_plot : bool
        If True, than a summary plot is updated in realtime during the run. ``sample_step`` defines the frequency at which the figure
        is updated.
    kwargs :
        Additional options defined in ``BaseFinder``

    """

    def __init__(
        self,
        parameter_set,
        metric="neg_mean_sq_err",
        sample_step=1,
        live_plot=False,
        **kwargs,
    ):
        super(PointEstimateFinder, self).__init__(parameter_set, **kwargs)
        self.metric_name = metric
        self.metric_func = get_metric(metric)
        self._weights = None
        self.sample_step = sample_step
        self.live_plot = live_plot

    def _get_evaluator(self, scm):
        assert (
            self.target is not None
        ), "set_target must be called before the fit can be performed"

        return Evaluator(self, scm)

    def find_x0(self, scm, method="random", **kwargs):
        """Explore the parameter space to find a good starting point to begin optimisation

        This is only required for ``scipy`` optimiser. The ``bayesopt`` optimiser does not require initial conditions.

        Parameters
        ----------
        scm: ``scmcallib.scm.BaseSCM``
            SCM instance to run
        method : str
            The method for exploring the tune parameter space. Currently only randomly spampling the parameter space has been
            implemented. (Default value = 'random')
        kwargs :
            Optional parameters to pass to the sampling method

            - 'random`

                samples : int
                    The number of random samples to perform

        Returns
        -------
        dict:
            A best guess for the Tune Parameters
        """
        evaluator = self._get_evaluator(scm)
        if method == "random":
            results = evaluator.run_multiple(
                self.parameter_set.evaluate(kwargs.get("samples", 100))
            )
        else:
            raise ValueError("Unknown search method")

        self._initial_guesses = results

        # Find the value with the largest metric_value
        row = results.loc[results["metric_value"].dropna().idxmax()]

        tune_params = {k: row[k] for k in self.parameter_set.names}
        logger.info(
            "found x0: {} tune_params: {}".format(row["metric_value"], tune_params)
        )
        return tune_params

    def find_best_fit(self, scm, include_bounds=True, optimiser_name="scipy", **kwargs):
        """Find the optimal point given a starting point

        Parameters
        ----------
        scm: ``scmcallib.scm.BaseSCM``
            SCM instance to run
        include_bounds :
            If True, use hard bounds which limits the algorithm choices. Otherwise, values outside
            the bounds are not passed to the model, but result in a large penalty
        optimiser_name :
            The optimiser to use. Options are ['scipy', 'bayesopt']
             (Default value = "scipy")
        kwargs :
            Additional arguments to pass to the optimiser.

        Returns
        -------
        EmulationResult

        See Also
        --------
        OptimiserBayesOpt
        OptimiserScipy

        """
        logger.info("Beginning search for best fit")
        e = self._get_evaluator(scm)
        optimiser = get_optimiser(optimiser_name)()

        res = optimiser.find_best_fit(
            e, self.parameter_set, include_bounds=include_bounds, **kwargs
        )

        if res.success:
            # Run the best value one more time as the last value is assumed to be the fit
            m = e(**res.x, force_add=True)
            logger.info(
                "converged with a best fit of {} with a metric value of {}".format(
                    res.x, m
                )
            )

        return EmulationResult(res, self.target, e)

    def set_target(self, observed, iter_over=None, weights=None, **kwargs):
        """Specify the target timeseries

        During optimisation the SCM output will be compared to these data.

        The ScmDataFrame must contain unique combinations of variable and regions. Other metadata are ignored, except in the case
        where `iter_over` is specified. In that case `observed` must contain unique combinations of 'variable', 'region' and the
        columns which are being iterated over.

        Parameters
        ----------
        observed : ScmDataFrame or arraylike
            Observational/model data to which the SCM should be calibrated. This can be an ScmDataFrame or an arraylike.
            If using an arraylike object, additional ``kwargs`` may be needed.
        iter_over: str or list of str:
            Specify which metadata values to iterate over allowing for multiple model configurations to be run per optimiser step.
        weights: dict or openscm.scmdataframe.ScmDataFrame
            Relative weighting of the samples. See examples below for usage
        kwargs :
            Other options to convert an array to a ScmDataFrame. See ``create_iam_dataframe`` for available options.

        Examples
        --------

        If a dict of weights is provided, the keys are used to apply weights to variables. The ``filter`` syntax of ScmDataFrame is
        used so '*' can be used to specify many variables.

        >>> observed = ScmDataFrame(np.arange(20).reshape((10, 2)), columns={
            "index": range(2000, 2010),
            "model": ["a_iam"],
            "climate_model": ["a_model"],
            "scenario": ["a_scenario"],
            "region": ["World"],
            "variable": ["Surface Temperature", "SLR_EXPANSION"],
            "unit": ["K", "mm"],
        })
        >>> point_finder.set_target(observed, weights={
            'Surface Temperature': 10.
        })
        >>> point_finder.weights
        variable            Surface Temperature SLR_EXPANSION
        region                            World         World
        time
        2000-01-01 00:00:00                10.0           1.0
        2001-01-01 00:00:00                10.0           1.0
        2002-01-01 00:00:00                10.0           1.0
        2003-01-01 00:00:00                10.0           1.0
        2004-01-01 00:00:00                10.0           1.0
        2005-01-01 00:00:00                10.0           1.0
        2006-01-01 00:00:00                10.0           1.0
        2007-01-01 00:00:00                10.0           1.0
        2008-01-01 00:00:00                10.0           1.0
        2009-01-01 00:00:00                10.0           1.0


        Notes
        -----
        If overlapping variables are specified via dict weights, the order of operations can not be guaranteed

        """
        super(PointEstimateFinder, self).set_target(
            observed, iter_over=iter_over, **kwargs
        )

        if weights is None:
            return

        weights_df = deepcopy(observed)
        weights_df._data.loc[:, :] = 1.0  # TODO: add ScmDataFrame.set_data

        if isinstance(weights, ScmDataFrameBase):
            w = weights_df.timeseries()
            for idx, r in weights.timeseries().iterrows():
                w.loc[idx] = r
            weights_df = weights_df.__class__(w)  # Recreate an ScmDataFrame
        elif isinstance(weights, pd.Series):
            pass
        elif isinstance(weights, dict):
            for k in weights:
                v = np.atleast_2d(
                    weights[k]
                )  # This makes sure that v is of shape (1, n)
                idx = weights_df.filter(variable=k).meta.index
                if len(idx) == 0:
                    raise ValueError(
                        "Could not set weights for variable {}: no such variable".format(
                            k
                        )
                    )
                weights_df._data[idx] = v.T
        else:
            raise ValueError("Cannot handle weights of type {}".format(type(weights)))

        self._weights = prepare_dataframe(weights_df, self.iter_over)

    @property
    def weights(self):
        """Weights used to scale the target and scm timeseries

        The result contains a subset of meta columns

        The weights are set in ``PointEstimateFinder.set_target``

        Returns
        -------
        pd.DataFrame or None
        """
        return deepcopy(self._weights)



from collections import Iterable
from datetime import datetime
from logging import getLogger

from dateutil import parser
from openscm.scmdataframe import ScmDataFrameBase

from scmcallib.scm import get_scm
from scmcallib.scm.base import BaseSCM
from scmcallib.utils import create_iam_dataframe, prepare_dataframe

logger = getLogger(__name__)


def _get_scm(scm, scm_kwargs, config_params):
    if isinstance(scm, BaseSCM):
        return scm

    scm_kwargs = {} if scm_kwargs is None else scm_kwargs
    return get_scm(scm)(**scm_kwargs, **config_params)


class BaseFinder(object):
    """Base class for the different methodologies for obtaining parameter sets to emulate a target timeseries

    Parameters
    ----------
    parameter_set :
        An instance of scmcallib.parameterset.ParameterSet which is used to define the constraints
        on the model. The free parameters will be updated to find the combination of parameters which best match the
        target data set using `set_target`.
    scm :
        The name of the SCM to generate parameters. Options are one of the following: ['ar5ir', 'magicc6',
        'magicc7']
    reference_period :
        A tuple of (start_year, end_year) or an integer representing the year to use as the
        reference period. All values used within the emulation are calculated as anomalies wrt this period. For tuples,
        is reference period is open, i.e. (2000, 2010) includes both 2000 and 2010 in the calculation of anomalies.
    scm_kwargs :
        A dictionary of kwargs to pass to the SCM instance

    """

    def __init__(self, parameter_set, reference_period=(2000, 2010)):
        self.parameter_set = parameter_set

        self._set_reference_period(reference_period)

        self.target = None
        self.extra_meta = None
        self.iter_over = tuple()
        self.iter_over_values = None

    def _set_reference_period(self, reference_period):
        if not isinstance(reference_period, Iterable):
            reference_period = (reference_period,)

        if len(reference_period) > 2:
            raise ValueError(
                "reference_period should be either one or two values which can be cast to a datetime"
            )

        def parse_datetime(v, first=True):
            if isinstance(v, datetime):
                return v
            elif isinstance(v, int):
                return datetime(v, 1, 1) if first else datetime(v, 12, 31)
            else:
                return parser.isoparse(v)

        if len(reference_period) == 2:
            self.reference_period = (
                parse_datetime(reference_period[0], first=True),
                parse_datetime(reference_period[1], first=False),
            )
        else:
            self.reference_period = (
                parse_datetime(reference_period[0], first=True),
                parse_datetime(reference_period[0], first=False),
            )

    def set_target(self, observed, iter_over=None, **kwargs):
        """Specify the target timeseries

        During optimisation the SCM output will be compared to these data.

        The ScmDataFrame must contain unique combinations of variable and regions. Other metadata are ignored, except in the case where
        `iter_over` is specified. In that case `observed` must contain unique combinations of 'variable', 'region' and the columns
        which are being iterated over.

        Parameters
        ----------
        observed : ScmDataFrame or arraylike
            Observational/model data to which the SCM should be calibrated. This can be an ScmDataFrame or an arraylike.
            If using an arraylike object, additional ``kwargs`` may be needed.
        iter_over: str or list of str:
            Specify which metadata values to iterate over allowing for multiple model configurations to be run per optimiser step.
        kwargs :
            Other options to convert an array to a ScmDataFrame. See ``create_iam_dataframe`` for available options.

        """
        if not isinstance(observed, ScmDataFrameBase):
            observed = create_iam_dataframe(observed, **kwargs)

        if iter_over is not None:
            iter_over = self.iter_over = (
                [iter_over] if isinstance(iter_over, str) else list(iter_over)
            )
            assert all([c in observed.meta for c in iter_over])

        df = prepare_dataframe(observed, self.iter_over)
        self.extra_meta = observed.meta.set_index(df.columns.names)
        if df.isna().any(axis=None):
            logger.warning("removing timesteps with NaN values from target")
            df = df.dropna()

        if iter_over:
            self.iter_over_values = observed.meta[iter_over].drop_duplicates()
            self.iter_over_values = self.iter_over_values.to_dict("records")
            logger.info("iterating over {} each fit".format(self.iter_over_values))
        ref_period = df.loc[slice(*self.reference_period)]

        if len(ref_period) == 0:
            raise ValueError(
                "No target values within the reference period. Check dates of the target timeseries"
            )
        ref_mean = ref_period.mean()
        if ref_mean.isna().any():
            raise ValueError(
                "NaN mean over the reference period. Check dates/values of the target timeseries"
            )
        self.target = df - ref_mean



from copy import deepcopy
from logging import getLogger
from os.path import join, dirname, exists, splitext

import pandas as pd
from openscm.scmdataframe import df_append, ScmDataFrame
from pymagicc.io import (
    convert_magicc_to_openscm_regions,
    convert_magicc7_to_openscm_variables,
    MAGICCData,
)
from scipy.io import loadmat

from .distributions import Uniform, Normal
from .finder import PointEstimateFinder
from .parameterset import ScenarioParameterSet

logger = getLogger(__name__)


def _lookup_param_col(df, param, model_code):
    try:
        v = df.set_index("purpose").at[param, model_code]
    except KeyError:
        return None
    return _maybe_as_float(v)


def _maybe_as_float(v):
    try:
        return float(v)
    except ValueError:
        return v


def _handle_simcap_varnames(p, fname):
    """
    Interpret variable/region data from simcap naming convention
    """
    if p["name"] == "CARBONCYCLE":
        res = {"variable": "CARBONCYCLE_" + p["colcode"], "region": "World"}
    else:
        if p["colcode"].startswith("BOX"):
            p["colcode"] = p["colcode"][3:]
        res = {
            "variable": p["name"],
            "region": convert_magicc_to_openscm_regions(p["colcode"]),
        }

    if "_WITHIN_" in fname:
        v, fname = fname.split("_WITHIN_")
        res["file_variable"] = v
    else:
        res["file_variable"] = p["name"]
    res["file"] = fname

    return res


def get_target(tuning_data):
    weights = []
    data = []

    for sce in tuning_data:
        for t in tuning_data[sce]:
            variable = t["file_variable"]
            err_msg = "Could not find data for scenario {}, variable {} and region {} in {}".format(
                sce, variable, t["region"], t["file"]
            )

            _, file_ext = splitext(t["file"].lower())
            if file_ext == ".mat":
                logger.error(".mat files are no longer supported")
                continue
            elif file_ext == ".mag":
                df = MAGICCData(t["file"])
                if variable in df["variable"].values:
                    df = df.filter(variable=variable, region=t["region"])
                else:
                    df = df.filter(
                        variable=convert_magicc7_to_openscm_variables(variable),
                        region=t["region"],
                    )
            elif file_ext == ".csv":
                df = ScmDataFrame(
                    pd.read_csv(t["file"])
                )  # read as a normal SCM Dataframe
                if len(df["scenario"].unique()) > 1:
                    # if there are more th
                    df = df.filter(scenario=sce)
                    if not len(df):
                        logger.error(err_msg)
                        continue
                else:
                    # If only one scenario is present assume that it is correct
                    if sce not in df["scenario"].values:
                        logger.error(
                            "Scenario {} was not found in {}. Assuming data from scenario".format(
                                sce, t["file"]
                            )
                        )
                if variable in df["variable"].values:
                    df = df.filter(variable=variable, region=t["region"])
                else:
                    df = df.filter(
                        variable=convert_magicc7_to_openscm_variables(variable),
                        region=t["region"],
                    )
            else:
                raise ValueError(
                    "Unknown file extension encountered: {}".format(file_ext)
                )

            if len(df):
                df["scenario"] = sce
                if "variable" in t:
                    df["variable"] = t["variable"]
                data.append(df)

                # Copy the df and setup the weights
                w = deepcopy(df)
                w._data.loc[:, :] = t["weight"] if "weight" in t else 1.0
                weights.append(w)
            else:
                logger.error(err_msg)

    return df_append(data), df_append(weights)


def read_csv_tuningcore(
    fname, data_dir=None, xtrascen_fname="XTRASCENS.mat", scensets_fname="SCENSETS.mat"
):
    """
    Read in a tuning core file
    Parameters
    ----------
    fname: str
        Name of the file to read
    data_dir: str
        The directory which contains the supplementary files which are referenced in the tuning core file. If nothing
        is specified, data_dir defaults to the same directory as the tuning core file.
    xtrascen_fname: str
        Filename of the xtrascen file
    scensets_fname: str
        Filename of the scensets file

    Returns
    -------
    A dict containing the information parsed from the tuning file
    """
    if data_dir is None:
        data_dir = dirname(fname)

    df = pd.read_csv(fname, sep=";", skiprows=4, header=None)
    res = {"runs": []}

    # We are referencing specific col/row numbers as that is how simcap also does it

    model_info = df.iloc[5:, 1:5]
    model_info.columns = pd.Index(["code", "name", "descr", "do_tuning"])

    parameters = df.iloc[:, 5:].T

    # Not needed, but helps with my sanity
    cols = ["name", "scenario", "colcode", "purpose", "description"] + [
        str(v) for v in model_info["code"]
    ]
    parameters.columns = pd.Index(cols)

    scenarios = parameters["scenario"].dropna().unique()

    if xtrascen_fname is not None and exists(join(data_dir, xtrascen_fname)):
        xtrascen = loadmat(
            join(data_dir, xtrascen_fname), struct_as_record=False, squeeze_me=True
        )["XTRASCENS"]
        xtrascen_keys = xtrascen._fieldnames
    else:
        xtrascen = None
        xtrascen_keys = []
        logger.warning("Not using xtrascen file")

    if scensets_fname is not None and exists(join(data_dir, scensets_fname)):
        scensets = loadmat(
            join(data_dir, scensets_fname), struct_as_record=False, squeeze_me=True
        )["SCENSETS"]
        scensets_keys = scensets._fieldnames
    else:
        scensets = None
        scensets_keys = []
        logger.warning("Not using scensets data")

    for _, m in model_info.iterrows():
        # Header info
        model = {
            "code": m["code"],
            "name": m["name"],
            "description": m["descr"],
            "do_tuning": True if m["do_tuning"] == "1" else False,
            "fixed_parameters": [],
            "free_parameters": [],
            "scenarios": [],
        }
        model_code = m["code"]
        scenario_info = {s: {"tuning_data": [], "parameters": []} for s in scenarios}

        for _, p in parameters.iterrows():
            if p["purpose"] == "FIXEDVALUE":
                # Fixed parameter
                model["fixed_parameters"].append(
                    {"name": p["name"], "value": _maybe_as_float(p[model_code])}
                )
            elif p["purpose"] == "STARTVALUE":
                # Free parameter
                parameter_cols = parameters[parameters["name"] == p["name"]]

                val = _maybe_as_float(p[model_code])
                if not isinstance(val, float):
                    raise ValueError("value for {} is not a float".format(p["name"]))

                parameter = {"name": p["name"], "value": val}

                v = _lookup_param_col(parameter_cols, "MIN", model_code)
                if v is not None:
                    parameter["min"] = v

                v = _lookup_param_col(parameter_cols, "MAX", model_code)
                if v is not None:
                    parameter["max"] = v

                v = _lookup_param_col(parameter_cols, "DISTRIBUTION", model_code)
                if v is not None:
                    parameter["distribution"] = v
                else:
                    parameter["distribution"] = "uniform"

                model["free_parameters"].append(parameter)
            elif p["purpose"] == "DATSOURCE":
                # Figure out the different file format which simcap specifies
                # Namely, interpret a variable and region from name and colcode
                d = _handle_simcap_varnames(p, p[model_code])
                d["file"] = join(data_dir, d["file"])
                parameter_cols = parameters[
                    (parameters["name"] == p["name"])
                    & (parameters["scenario"] == p["scenario"])
                    & (parameters["colcode"] == p["colcode"])
                ]
                weight = _lookup_param_col(parameter_cols, "WEIGHT", model_code)
                if weight is not None:
                    d["weight"] = weight
                scenario_info[p["scenario"]]["tuning_data"].append(d)
            elif p["purpose"] == "XTRASCENSETTING":
                sce_params = {}

                if scensets is not None:
                    scenset_sce_name = "SCENSET_" + p["scenario"]
                    if scenset_sce_name in scensets_keys:
                        xtrascen_struct = getattr(scensets, scenset_sce_name)
                        params = {
                            f: getattr(xtrascen_struct, f)
                            for f in xtrascen_struct._fieldnames
                        }
                        sce_params = {**sce_params, **params}
                    else:
                        logger.error(
                            "Could not find scenset values for {}".format(p["value"])
                        )
                # Amend/overwrite with data from xtrascen
                if xtrascen is not None:
                    if p[model_code] in xtrascen_keys:
                        xtrascen_struct = getattr(xtrascen, p[model_code])
                        params = {
                            f: getattr(xtrascen_struct, f)
                            for f in xtrascen_struct._fieldnames
                        }
                        sce_params = {**sce_params, **params}
                    else:
                        logger.error(
                            "Could not find xtrascen values for {}".format(p["value"])
                        )
                scenario_info[p["scenario"]]["parameters"] = sce_params

        # Flatten the scenarios object
        model["scenarios"] = [
            {
                "name": s,
                "tuning_data": scenario_info[s]["tuning_data"],
                "parameters": scenario_info[s]["parameters"],
            }
            for s in scenario_info
        ]
        res["runs"].append(model)

    return res


def create_point_model(model, **kwargs):
    """

    Create a new model point estimate model instance from a configuration dict

    Note that a tuning core file may specify multiple calibration models. This function expects a single dict

    Parameters
    ----------
    model : dict
        Required keys (fixed_parameters, free_parameters)

        scenario: List of scenario definitions
            name : Name of the scenario
            parameters: List of fixed parameters for a given scenario
            tuning_data: List of scenario specific tuning_data
    Returns
    -------
    An instance of a PointEstimateFinder which has its parameter's and target's set

    """

    assert isinstance(model, dict)

    required_keys = ["fixed_parameters", "free_parameters"]
    for k in required_keys:
        if k not in model:
            raise ValueError('The model dict didnt contain a "{}" key'.format(k))
    scenarios = model.get("scenarios", [])

    params = ScenarioParameterSet([s["name"] for s in scenarios])

    for p in model["fixed_parameters"]:
        params.set_config(p["name"], p["value"])

    for p in model["free_parameters"]:
        if p["distribution"] == "uniform":
            dist = Uniform(lower=p["min"], upper=p["max"])
        elif p["distribution"] == "normal":
            dist = Normal(mu=p["mean"], sd=p["sd"])
        else:
            raise ValueError("Unknown distribution {}".format(p["distribution"]))
        params.set_tune(p["name"], dist, x0=p.get("value", None))

    # Find all the scenario parameters used
    sce_param_names = set()
    for sce in scenarios:
        for p in sce["parameters"]:
            sce_param_names.add(p)

    for p in sce_param_names:
        items = {sce["name"]: sce["parameters"].get(p, None) for sce in scenarios}
        params.set_config(p, items)

    m = PointEstimateFinder(params, **kwargs)

    # Load the data
    data, weights = get_target({sce["name"]: sce["tuning_data"] for sce in scenarios})
    m.set_target(data, iter_over="scenario", weights=weights)

    return m



from concurrent.futures import ProcessPoolExecutor, as_completed, wait
from copy import deepcopy
from datetime import datetime
from logging import getLogger

import numpy as np
import pandas as pd
from openscm.scmdataframe import ScmDataFrame
from tqdm import tqdm

from . import mat4py

logger = getLogger(__name__)


# ZN: can we get rid of most of these given the new capabilities of ScmDataFrame?
def load_iam_timeseries(fname):
    # Read in the data. It requires a bit of wrangling to get it in the IamDataFrame format
    df = pd.read_csv(fname, header=None)
    df = df.set_index(0).T
    df.pop("time")
    df = ScmDataFrame(df)
    df["time"] = [
        datetime(d.year, d.month, d.day) for d in df["time"]
    ]  # strip out any hour values
    df = df.filter(
        region=[
            "World",
            "World|Northern Hemisphere|Land",
            "World|Northern Hemisphere|Ocean",
            "World|Southern Hemisphere|Land",
            "World|Southern Hemisphere|Ocean",
        ]
    )

    # Resample to annual means
    return df.resample("AS")


def merge_iam_timeseries(*args):
    """Merges two or more timeseries in the ScmDataFrame format.

    This method assumes that all data are from the same scenario. All scenarios in the resulting dataframe will be the same as specified
    by data in the last value of ``dfs``

    Parameters
    ----------
    args : openscm.scmdataframe.ScmDataFrame or string
        2 or more items to merge. ``openscm.scmdataframe.ScmDataFrame`` objects can be passed directly otherwise, if a string is passed
        then the dataframe will be attempted to be read from file

    Returns
    -------
    openscm.scmdataframe.ScmDataFrame

    """
    dfs = args
    assert len(dfs) >= 2, "At least two input dfs are required"

    def get_df(f):
        if isinstance(f, ScmDataFrame):
            return f
        else:
            return load_iam_timeseries(f)

    # Ensure that each df is loaded
    dfs = [get_df(f) for f in dfs]
    target_scenario = dfs[-1]["scenario"].unique()[0]
    res = None
    for df in dfs:
        df_copy = deepcopy(df)

        df_copy["scenario"] = target_scenario
        if res is None:
            res = df_copy
        else:
            res = res.append(df_copy)

    return res


def create_iam_dataframe(
    ts,
    index=None,
    variable="Surface Temperature",
    scenario="unspecified",
    model="unspecified",
    region="World",
    unit="K",
):
    """Create a IamDataFrame from a timeseries

    A number of the interfaces within scmcallib expect IamDataFrames, but they are not always the easiest things to initialise. This
    function provides a helper function for converting a 1d timeseries into a IamDataFrame

    Parameters
    ----------
    ts :
        1D np array or pd.Series/pd.Dataframe.
    index :
        If None, use the index from the ts. This assumes that `ts` is a pd.Series or pd.DataFrame. (Default value = None)
    variable :
         (Default value = 'Surface Temperature')
    scenario :
         (Default value = 'unspecified')
    model :
         (Default value = 'unspecified')
    region :
         (Default value = 'World')
    unit :
         (Default value = 'K')

    Returns
    -------
    type
        openscm.scmdataframe.ScmDataFrame instance

    """
    if index is None:
        if isinstance(ts, (pd.DataFrame, pd.Series)):
            index = ts.index
        else:
            raise ValueError("No valid index")

    if isinstance(ts, (pd.DataFrame, pd.Series)):
        assert len(ts)
        ts = ts.values.squeeze()

    assert ts.ndim == 1
    return ScmDataFrame(
        pd.Series(ts, index=index),
        columns={
            "unit": [unit],
            "variable": [variable],
            "region": [region],
            "model": [model],
            "scenario": [scenario],
        },
    )


def prepare_dataframe(df, extra_cols=()):
    """Convert model output in the form of a ScmDataFrame into something more useful in this use case

    Parameters
    ----------
    df : openscm.scmdataframe.ScmDataFrame
        Timeseries data from an scm or other training data
    extra_cols : listlike of str
        Additional meta columns to include in the output DataFrame

    Returns
    -------
    pd.DataFrame
        DataFrame with a time index and MultiIndex columns for model and region

    """
    return df.timeseries(meta=["variable", "region"] + list(extra_cols)).T


def convert_tuningstruc_to_scmdf(
    filepath, variable=None, region=None, unit=None, scenario=None, model=None
):
    """Convert a matlab tuningstruc to an ScmDataFrame

    Parameters
    ----------
    filepath : str
        Filepath from which to load the data

    variable : str
        Name of the variable contained in the tuningstruc. If None,
        `convert_tuningstruc_to_scmdf` will attempt to determine it from the input
        file.

    region : str
        Region to which the data in the tuningstruc applies. If None,
        `convert_tuningstruc_to_scmdf` will attempt to determine it from the input
        file.

    unit : str
        Units of the data in the tuningstruc. If None,
        `convert_tuningstruc_to_scmdf` will attempt to determine it from the input
        file.

    scenario : str
        Scenario to which the data in the tuningstruc applies. If None,
        `convert_tuningstruc_to_scmdf` will attempt to determine it from the input
        file.

    model : str
        The (integrated assessment) model which generated the emissions scenario
        associated with the data in the tuningstruc. If None,
        `convert_tuningstruc_to_scmdf` will attempt to determine it from the input
        file and if it cannot, it will be set to "unspecified".

    Raises
    ------
    ValueError
        If a metadata variable is not supplied and it cannot be determined from the
        tuningstruc.

    Returns
    -------
    :obj: `ScmDataFrame`
        ScmDataFrame with the tuningstruc data
    """
    dataset = mat4py.loadmat(filepath)

    for m, climate_model in enumerate(dataset["tuningdata"]["modelcodes"]):
        metadata = {
            "variable": [variable],
            "region": [region],
            "unit": [unit],
            "climate_model": [climate_model],
            "scenario": [scenario],
            "model": [model],
        }
        for k, v in metadata.items():
            if v == [None]:
                try:
                    metadata[k] = [dataset["tuningdata"]["model"][m][k]]
                except KeyError:
                    if k == "model":
                        metadata[k] = ["unspecified"]
                        continue

                    error_msg = "Cannot determine {} from file: " "{}".format(
                        k, filepath
                    )
                    raise KeyError(error_msg)

        scmdf = ScmDataFrame(
            data=np.asarray(dataset["tuningdata"]["model"][m]["data"][1]),
            columns=metadata,
            index=dataset["tuningdata"]["model"][m]["data"][0],
        )
        try:
            ref_df.append(scmdf, inplace=True)
        except NameError:
            ref_df = scmdf

    return ref_df


def convert_scmdf_to_tuningstruc(scmdf, outpath):
    """Convert an ScmDataFrame to a matlab tuningstruc

    One tuningstruc file will be created for each unique
    ["model", "scenario", "variable", "region", "unit"] combination in the input
    ScmDataFrame.

    Parameters
    ----------
    scmdf : :obj: `ScmDataFrame`
        ScmDataFrame to convert to a tuningstruc

    outpath : str
        Base path in which to save the tuningstruc. The rest of the pathname is
        generated from the metadata. `.mat` is also appended automatically.
    """
    iter = scmdf.timeseries().groupby(
        ["model", "scenario", "variable", "region", "unit"]
    )
    for (model, scenario, variable, region, unit), df in iter:
        dataset = {}
        dataset["tuningdata"] = {}
        dataset["tuningdata"]["modelcodes"] = []
        dataset["tuningdata"]["model"] = []

        for m, (climate_model, df) in enumerate(df.groupby("climate_model")):
            # impossible to make dataframe with duplicate rows so not an issue
            # this is just in case
            error_msg = (
                "Should only have a single unique timeseries for a given "
                '["climate_model", "model", "scenario", "variable", '
                '"region", "unit"] combination'
            )
            assert df.shape[0] == 1, error_msg

            dataset["tuningdata"]["modelcodes"].append(climate_model)
            dataset["tuningdata"]["model"].append({})

            dataset["tuningdata"]["model"][m]["model"] = model
            dataset["tuningdata"]["model"][m]["scenario"] = scenario
            dataset["tuningdata"]["model"][m]["variable"] = variable
            dataset["tuningdata"]["model"][m]["region"] = region
            dataset["tuningdata"]["model"][m]["unit"] = unit

            dataset["tuningdata"]["model"][m]["notes"] = (
                "{} {} {} {} ({}) tuningstruc (written with scmcallib)"
                "".format(scenario, model, region, variable, unit)
            )
            dataset["tuningdata"]["model"][m]["data"] = [
                [float(t.year) for t in df.columns],
                list(df.values.squeeze()),
            ]
            dataset["tuningdata"]["model"][m]["col_code"] = ["YEARS", variable]

        outfile = (
            "{}_{}_{}_{}_{}.mat".format(outpath, scenario, model, variable, region)
            .replace(" ", "_")
            .replace("|", "_")
        )
        # ask Jared how to add logging here
        mat4py.savemat(outfile, dataset)


def parallel_process(
    func, arr, n_jobs=10, use_kwargs=False, front_num=3, show_progress=False, pool=None
):
    """
    A parallel version of the map function with a progress bar.

    Adapted from http://danshiebler.com/2016-09-14-parallel-progress-bar/

    Parameters
    ----------
    func : function
        A python function to apply to the elements of array
    arr: listlike
        An array of configuration to iterate over
    n_jobs: int default=16
    use_kwargs : boolean, default=False
        Whether to consider the elements of array as dictionaries of keyword arguments to function
    front_num : int, default=3
        The number of iterations to run serially before kicking off the parallel job. Useful for catching bugs

    Returns
    -------
        [func(arr[0]), func(arr[1]), ...]
    """

    # We run the first few iterations serially to catch bugs
    front = []
    if front_num > 0:
        front = [func(**a) if use_kwargs else func(a) for a in arr[:front_num]]
    # If we set n_jobs to 1, just run a list comprehension. This is useful for benchmarking and debugging.
    if n_jobs == 1:
        if show_progress:
            return front + [
                func(**a) if use_kwargs else func(a) for a in tqdm(arr[front_num:])
            ]
        else:
            return front + [
                func(**a) if use_kwargs else func(a) for a in arr[front_num:]
            ]

    cleanup_pool = False
    if pool is None:
        pool = ProcessPoolExecutor(n_jobs)
        cleanup_pool = True

    try:
        # Pass the elements of array into func
        if use_kwargs:
            futures = [pool.submit(func, **a) for a in arr[front_num:]]
        else:
            futures = [pool.submit(func, a) for a in arr[front_num:]]
        kwargs = {
            "total": len(futures),
            "unit": "it",
            "unit_scale": True,
            "leave": True,
        }
        # Print out the progress as tasks complete
        if show_progress:
            for f in tqdm(as_completed(futures), **kwargs):
                pass
        else:
            wait(futures)
    finally:
        if cleanup_pool:
            pool.shutdown()

    out = []
    # Get the results from the futures.
    for i, future in enumerate(futures):
        try:
            out.append(future.result())
        except Exception as e:
            out.append(e)
    return front + out



from .scipy import OptimiserScipy
from .bayesopt import OptimiserBayesOpt

_optimisers = [OptimiserScipy, OptimiserBayesOpt]


def get_optimiser(optimiser_name):
    for o in _optimisers:
        if o.name == optimiser_name.lower():
            return o

    raise ValueError(
        "No optimiser named {}. Options include {}".format(
            optimiser_name, [m.name for m in _optimisers]
        )
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from os.path import realpath, dirname

from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand

import versioneer

PROJECT_ROOT = dirname(realpath(__file__))

PACKAGE_NAME = "scmcallib"
DESCRIPTION = "Framework for calibrating simple climate models"
KEYWORDS = [
    "calibration",
    "tuning",
    "simple climate model",
    "reduced complexity climate model",
    "data processing",
]
AUTHORS = ["Jared Lewis", "Zebedee Nicholls", "Matthias Mengel"]

AUTHOR_EMAILS = [
    "jared.lewis@climate-energy-college.org",
    "zebedee.nicholls@climate-energy-college.org",
    "matthias.mengel@pik-potsdam.de",
]
URL = "https://gitlab.com/magicc/scmcallib"
LICENSE = "2-Clause BSD License"
CLASSIFIERS = [
    "Development Status :: 3 - Alpha",
    "Intended Audience :: Science/Research",
    "Natural Language :: English",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.7",
    "Topic :: Scientific/Engineering :: Atmospheric Science",
]

REQUIREMENTS_INSTALL = [
    "pymc3==3.7",
    "numpy",
    "pyam-iamc>=0.2.0",
    "pandas==0.24.2",  # pyam issue, see https://github.com/IAMconsortium/pyam/issues/245
    "matplotlib>=2.0.0",
    "bayesian-optimization",
    "openscm>=0.1.0a",
    "pymagicc>=2.0.0b4",
    "scikit-learn",
]
REQUIREMENTS_TESTS = ["codecov", "pytest-cov", "pytest>=4.0,<5.0"]
REQUIREMENTS_NOTEBOOKS = ["notebook", "nbval", "expectexception"]
REQUIREMENTS_DOCS = ["sphinx>=1.4,<2.1", "sphinx_rtd_theme", "sphinx-click"]
REQUIREMENTS_DEPLOY = ["twine>=2.0.0", "setuptools>=38.6.0", "wheel>=0.31.0"]
REQUIREMENTS_DEV = [
    *["flake8>=3.7.0", "black"],
    *REQUIREMENTS_TESTS,
    *REQUIREMENTS_NOTEBOOKS,
    *REQUIREMENTS_DOCS,
    *REQUIREMENTS_DEPLOY,
]

REQUIREMENTS_EXTRAS = {
    "docs": REQUIREMENTS_DOCS,
    "tests": REQUIREMENTS_TESTS,
    "notebooks": REQUIREMENTS_NOTEBOOKS,
    "deploy": REQUIREMENTS_DEPLOY,
    "dev": REQUIREMENTS_DEV,
}
with open("README.rst", "r") as readme_file:
    README_TEXT = readme_file.read()


class Scmcallib(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import pytest

        pytest.main(self.test_args)


cmdclass = versioneer.get_cmdclass()
cmdclass.update({"test": Scmcallib})

setup(
    name=PACKAGE_NAME,
    keywords=KEYWORDS,
    url=URL,
    author=",".join(AUTHORS),
    author_email=",".join(AUTHOR_EMAILS),
    description="Perform calibration for simple climate models",
    long_description=README_TEXT,
    long_description_content_type="text/x-rst",
    classifiers=CLASSIFIERS,
    include_package_data=True,
    packages=find_packages(include=["scmcallib"]),
    test_suite="tests",
    install_requires=REQUIREMENTS_INSTALL,
    extras_require=REQUIREMENTS_EXTRAS,
    version=versioneer.get_version(),
    cmdclass=cmdclass,
    zip_safe=False,
)



# -*- coding: utf-8 -*-

"""Top-level package for SCM Calibration."""

import os
import sys

# Order matters here. MKL_THREADING_LAYER has to be set prior to any theano calls
os.environ["MKL_THREADING_LAYER"] = "GNU"

# Disable the theano debug hook. It messes with pycharm stack traces and won't be needed by scmcallib
from theano.gof.link import __excepthook

sys.excepthook = __excepthook

from ._version import get_versions
from .finder import PointEstimateFinder, DistributionFinder
from .distributions import Normal, LogNormal, Scalar, Bound
from .parameterset import ParameterSet

__version__ = get_versions()["version"]
del get_versions
